"""
Implementation of the `Either` type for handling computation results with two possible types.

The `Either` type represents a value that can be one of two types: `Left` or `Right`.
Typically used for computations that can either succeed (`Right`) or fail (`Left`).
"""

from collections.abc import Callable
from typing import Any, NoReturn


class Left[TLeft]:
    """
    The left variant of an `Either` type.

    Typically represents the failure case in an `Either` computation.
    """

    def __init__(self, value: TLeft) -> None:
        """
        Initialize a `Left` value.

        Args:
            value: The value to store.
        """
        super().__init__()
        self.value = value

    @staticmethod
    def is_left() -> bool:
        """
        Check if this is a `Left` value.

        Returns:
            Always `True` for `Left` instances.
        """
        return True

    @staticmethod
    def is_right() -> bool:
        """
        Check if this is a `Right` value.

        Returns:
            Always `False` for `Left` instances.
        """
        return False

    def get_left_or_raise(self) -> TLeft:
        """
        Get the contained `Left` value.

        Returns:
            The contained `Left` value.
        """
        return self.value

    @staticmethod
    def get_right_or_raise() -> NoReturn:
        """
        Attempt to get a `Right` value.

        Raises:
            ValueError: Always raises since this is a `Left` instance.
        """
        raise ValueError("Value is left")

    def map[TReturnLeft](
        self,
        left: Callable[[TLeft], TReturnLeft],
        right: Callable[[Any], Any],  # noqa: ARG002
    ) -> TReturnLeft:
        """
        Transform the contained value.

        Args:
            left: Function to transform `Left` values.
            right: Function to transform `Right` values (unused).

        Returns:
            Result of applying the left function to the value.
        """
        return left(self.value)


class Right[TRight]:
    """
    The right variant of an `Either` type.

    Typically represents the success case in an `Either` computation.
    """

    def __init__(self, value: TRight) -> None:
        """
        Initialize a `Right` value.

        Args:
            value: The value to store.
        """
        super().__init__()
        self.value = value

    @staticmethod
    def is_left() -> bool:
        """
        Check if this is a `Left` value.

        Returns:
            Always `False` for `Right` instances.
        """
        return False

    @staticmethod
    def is_right() -> bool:
        """
        Check if this is a `Right` value.

        Returns:
            Always `True` for `Right` instances.
        """
        return True

    @staticmethod
    def get_left_or_raise() -> NoReturn:
        """
        Attempt to get a `Left` value.

        Raises:
            ValueError: Always raises since this is a `Right` instance.
        """
        raise ValueError("Value is right")

    def get_right_or_raise(self) -> TRight:
        """
        Get the contained `Right` value.

        Returns:
            The contained `Right` value.
        """
        return self.value

    def map[TReturnRight](
        self,
        left: Callable[[Any], Any],  # noqa: ARG002
        right: Callable[[TRight], TReturnRight],
    ) -> TReturnRight:
        """
        Transform the contained value.

        Args:
            left: Function to transform `Left` values (unused).
            right: Function to transform `Right` values.

        Returns:
            Result of applying the right function to the value.
        """
        return right(self.value)


type Either[TLeft, TRight] = Left[TLeft] | Right[TRight]
"""
Implementation of the `Either` type for handling computation results with two possible types.

The `Either` type represents a value that can be one of two types: `Left` or `Right`.
Typically used for computations that can either succeed (`Right`) or fail (`Left`).
"""
